# Copyright (C) 2005, Timothy R. Morgan
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice and this permission
# notice appear in all copies, modified copies and in
# supporting documentation.

"React Model"

__version__ = '0.1'

from Globals import InitializeClass, DTMLFile
from OFS.SimpleItem import SimpleItem
from AccessControl import ClassSecurityInfo
from OFS.PropertyManager import PropertyManager
from Shared.DC.ZRDB.Results import Results

class ReactModel(SimpleItem, PropertyManager):
  "React Model"

  meta_type = 'React Model'
  
  security = ClassSecurityInfo()
  
  manage_options = (
    PropertyManager.manage_options
    + SimpleItem.manage_options
  )

  _properties = (
    {'id': 'db_id', 'type': 'string', 'mode': 'w'},
    {'id': 'db_type', 'type': 'selection', 'mode': 'w', 'select_variable':'db_types'},
  )
  
  db_types = ['mysql', 'mssql', 'foxpro']
  
  def __init__(self, id, db_id, db_type):
    self.id = id
    self.db_id = db_id
    self.db_type = db_type
  
  def get_db(self):
    try:
      db = getattr(self.aq_parent, self.db_id)
    except:
      if self.db_id:
        raise Exception, 'Could not find database connection object "%s"' % self.db_id
      else:
        raise Exception, 'No database connection object specified.'
    return db._v_database_connection
  
  def new(self, **kwargs):
    return ReactModelRecord(self.get_db(), self.db_id, self.db_type, self.id, **kwargs)

  def find_all(self, conditions=None):
    if conditions and type(conditions) == type([]):
      conditions = ' and '.join(conditions)
    if conditions:
      query = 'select id from %s where %s' % (self.id, conditions)
    else:
      query = 'select id from %s' % self.id
    db = self.get_db()
    # this is inefficient, but will work for now
    records = []
    for result in db.query(query)[1]:
      id = result[0]
      record = ReactModelRecord(self.get_db(), self.db_id, self.db_type, self.id)
      record.load_by_sql('id = %s' % id)
      records.append(record)
    return records
  
  def find(self, id):
    record = ReactModelRecord(self.get_db(), self.db_id, self.db_type, self.id)
    sql = 'id = %s' % id
    record.load_by_sql(sql)
    return record
  
  def find_by_sql(self, sql):
    record = ReactModelRecord(self.get_db(), self.db_id, self.db_type, self.id)
    record.load_by_sql(sql)
    return record
  
  def find_by_value(self, column, value):
    record = ReactModelRecord(self.get_db(), self.db_id, self.db_type, self.id)
    sql = "%s = %s" % (column, record._quote_value(value))
    record.load_by_sql(sql)
    return record


InitializeClass(ReactModel)

manage_addReactModelForm = DTMLFile('www/ReactModelAdd', globals())

def manage_addReactModel(self, id, db_id, db_type, REQUEST=None, submit=None):
  """Adds a new ReactModel object."""
  self._setObject(id, ReactModel(id, db_id, db_type))
  if REQUEST != None:
    REQUEST.RESPONSE.redirect(REQUEST['URL1']+'/manage_main')


class ReactModelRecord(object):
  """
  Acts as a proxy object represeting a database record.
  Attributes on the object represent columns of the record,
  however changes to the attributes are not written back to
  the database in real-time; the 'save' method does that.
  """
  
  # If a Zope script/dtml/whatever object can get ahold of
  # one of these objects, then it should be able to do
  # whatever it wants, so give it everything.
  __allow_access_to_unprotected_subobjects__ = 1
  _guarded_writes = 1
  
  def __init__(self, db, db_id, db_type, table, **kwargs):
    self._db = db
    self._db_id = db_id
    self._db_type = db_type
    self._table = table
    self.id = None
    # keyword arguments are treated as columns of the table
    # and added to the object
    for key in kwargs.keys():
      setattr(self, key, kwargs[key])
  
  def __guarded_setitem__(self, key, value):
    """
    allows setting of normal attributes (no underscores please)
    """
    if not key.startswith('_') and key != 'id':
      setattr(self, key, value)
  
  def load_by_sql(self, sql):
    """
    query with the sql provided, get the first record from the
    results, and append the columns as attributes to the object
    """
    try:
      results = self._db.query('select * from %s where %s' % (self._table, sql))
      record = Results(results).dictionaries()[0]
    except:
      raise Exception, 'Record could not be found.'
    else:
      for key in record.keys():
        setattr(self, key, record[key])
  
  def save(self):
    """
    Write all the object's attributes to the database table.
    If this object has an id, then do an update to an existing
    record; if not, insert a new record.
    """
    dict = self.__dict__.copy()
    for key in dict.keys():
      if key.startswith('_') or key == 'id':
        del(dict[key])
    columns = dict.keys()
    # UPDATE
    if self.id:
      sets = []
      for column in columns:
        value = self._quote_value(dict[column])
        sets.append('%s = %s' % (column, value))
      query = 'update %s set %s where id = %s' % (self._table, ', '.join(sets), self.id)
    # INSERT
    else:
      values = []
      for column in columns:
        value = self._quote_value(dict[column])
        values.append(value)
      query = 'insert into %s (%s) values (%s)' % (self._table, ', '.join(columns), ', '.join(values))
    self._db.query(query)
    # get the id (after insert)
    if not self.id:
      # Getting the auto_increment (or identity as mssql calls it) id
      # is very database-specific. That's the main reason why we need to
      # know what the database type is.
      if self._db_type == 'mysql':
        self.id = self._db.query('select LAST_INSERT_ID()')[1][0][0]
      elif self._db_type == 'mssql':
        self.id = self._db.query('select @@IDENTITY')[1][0][0]
      else:
        # This is pretty ugly, but it should work in most situations.
        # This assumes that the combination of all the values (other than the id)
        # is unique. But even if there's another record with all the same
        # values, wouldn't it be OK to retrieve the id of either record?
        wheres = []
        for column in columns:
          value = self._quote_value(dict[column])
          wheres.append('%s = %s' % (column, value))
        query = "select id from %s where %s" % (self._table, ' and '.join(wheres))
        self.id = self._db.query(query)[1][-1][0]
  
  def destroy(self):
    """
    Delete the record from the database.
    """
    if self.id:
      query = 'delete from %s where id = %s' % (self._table, self.id)
      self._db.query(query)

  def _quote_value(self, value):
    """
    Return the properly stringified and quoted value based
    on the database type.
    """
    if type(value) == type(''):
      value = quote(value)
    elif hasattr(value, 'strftime'):
      if self._db_type == 'foxpro':
        value = quote(value.strftime('DATETIME(%Y,%m,%d,%H,%M,%S)'))
      else:
        value = quote(value.strftime('%Y/%m/%d %H:%M:%S'))
    else:
      value = str(value)
    return value

def quote(value):
  """
  Escape quotes for a SQL query.
  """
  return "'" + value.replace("'", "''") + "'"
